#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <windows.h>
#include <ctype.h>

/*  max number of emails */
#define MAX 100000

#define MAXS 64
typedef struct {
  char path[MAXS];
  char from[MAXS];
  char subject[MAXS];
  char dato[MAXS];
} mail_t;

mail_t mails[MAX];
int n;

int compm(const void *A,const void *B) {
  const mail_t *a=A,*b=B;
  int r;
  if((r=strcmp(a->dato,b->dato))) return r;
  if((r=strcmp(a->from,b->from))) return r;
  if((r=strcmp(a->subject,b->subject))) return r;
/*  printf("[mail at %s from %s [%s] in %d exists twice]\n",a->path,a->from,a->subject,a->y);*/
  return 0;
}

void usage() {
  puts("tcomp by ruben spaans in december 2011, compare two thunderbird profiles\n");
  puts("  usage: tcomp dir1 dir2\n");
  puts("the output is a list of emails contained in dir1 and not in dir2.\n");
  puts("example: tcomp c:/apps/thunderbird/mail d:/backup/mail");
  exit(0);
}

int good(char *s) {
  int l=strlen(s);
  if(!l) return 0;
  return s[l-1]=='/' || s[l-1]==':' || s[l-1]=='\\';
}

void kopier(char *to,char *from) {
  int i=MAXS-1;
  while(i-- && *from) *to++=*from++;
  *to=0;
}

void extractdate(char *t,int *has,mail_t *mail) {
  if(strncmp(t,"Date: ",6)) return;
  kopier(mail->dato,t+6);
  *has=1;
}

void extractfrom(char *t,int *has,mail_t *mail) {
  if(strncmp(t,"From: ",6)) return;
  kopier(mail->from,t+6);
  *has=1;
}

void extractsubj(char *t,int *has,mail_t *mail) {
  if(strncmp(t,"Subject: ",9)) return;
  kopier(mail->subject,t+9);
  *has=1;
}

int isstart(char *t) {
  return !strncmp(t,"From - ",7);
}

#define MAXLINE 131072
void parsefile(char *path,char *filename,void (*func)(mail_t *),int out) {
  static char s[1024],t[MAXLINE+1];
  FILE *f;
  mail_t mail;
  int hasdate=0,hasfrom=0,hassubj=0,z,num=0,start=0;
  strcpy(s,path);
  strcat(s,filename);
  if(!(f=fopen(s,"r"))) {
    printf("[couldn't open %s]\n",s);
    return;
  }
  kopier(mail.path,s);
#define GET if(!fgets(t,MAXLINE,f)) break
  while(1) {
    GET;
    if(isstart(t)) start=1;
    if(!start) continue;
    z=strlen(t); if(t[z-1]=='\n') t[z-1]=0;
    if(!hasdate) extractdate(t,&hasdate,&mail);
    if(!hasfrom) extractfrom(t,&hasfrom,&mail);
    if(!hassubj) extractsubj(t,&hassubj,&mail);
    if(hasdate&hasfrom&hassubj) {
      func(&mail);
      hasdate=hasfrom=hassubj=start=0;
      num++;
    }
  }
  if(out)printf("[found %d emails in %s]\n",num,s);
  fclose(f);
}
#undef GET
#undef MAXLINE

void add(mail_t *elem) {
  mails[n++]=*elem;
}

void find(mail_t *elem) {
  int lo=0,hi=n,mid,r;
  while(lo!=hi) {
    mid=lo+(hi-lo)/2;
    r=compm(elem,mails+mid);
    if(r>0) lo=mid+1;
    else hi=mid;
  }
  if(compm(elem,mails+lo)) {
    /*  not found */
    printf("[-------mail not found in second profile--------]\n");
    printf("   date [%s]\n",elem->dato);
    printf("   from [%s]\n",elem->from);
    printf("subject [%s]\n",elem->subject);
    printf("at path [%s]\n",elem->path);
  }
}

void check(char *path,void (*func)(mail_t *),int out) {
  HANDLE hfind;
  WIN32_FIND_DATA f;
  char t[1024],u[1024];
  strcpy(t,path);
  strcat(t,"*");
  hfind=FindFirstFile(t,&f);
  if(INVALID_HANDLE_VALUE==hfind) {
    printf("[error reading dir %s]\n",t);
    return;
  }
  do {
    if(f.dwFileAttributes&FILE_ATTRIBUTE_DIRECTORY) {
      /*  skip . and .. */
      if(!strcmp(f.cFileName,".")) continue;
      if(!strcmp(f.cFileName,"..")) continue;
      strcpy(u,path);
      strcat(u,f.cFileName);
      strcat(u,"/");
      check(u,func,out);
    } else {
      int l=strlen(f.cFileName);
      if(l>3 && f.cFileName[l-4]=='.' && f.cFileName[l-3]=='s' && f.cFileName[l-2]=='b' && f.cFileName[l-1]=='d') continue;
      if(l>3 && f.cFileName[l-4]=='.' && f.cFileName[l-3]=='m' && f.cFileName[l-2]=='s' && f.cFileName[l-1]=='f') continue;
      parsefile(path,f.cFileName,func,out);
    }
  } while(FindNextFile(hfind,&f));
  FindClose(hfind);
}

int main(int argc,char **argv) {
  char path[1024];
  int i,j;
  if(argc<3) usage();
  n=0;
  strcpy(path,argv[2]);
  if(!good(path)) strcat(path,"/");
  check(path,add,1);
  printf("[found %d emails]\n",n);
  qsort(mails,n,sizeof(mail_t),compm);
  /*  remove duplicates */
  for(i=j=1;i<n;i++) if(compm(mails+i,mails+i-1)) mails[j++]=mails[i];
  n=j;
  printf("[emails reduced to %d due to duplicates]\n",n);
  strcpy(path,argv[1]);
  if(!good(path)) strcat(path,"/");
  check(path,find,1);
  return 0;
}
